#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import yaml
from typing import Union
from disdrodb.l0.io import (
    get_disdrodb_dir,
    get_data_source,
    get_campaign_name,
)
from disdrodb.api.metadata import get_metadata_list, read_station_metadata
from disdrodb.l0.l0_reader import _check_metadata_reader
from disdrodb.l0.metadata import (
    _check_metadata_keys,
    _check_metadata_data_source,
    _check_metadata_campaign_name,
    _check_metadata_station_name,
    _check_metadata_sensor_name,
    check_metadata_compliance,
)

#### --------------------------------------------------------------------------.


def read_yaml(fpath: str) -> dict:
    """Read YAML file.

    Parameters
    ----------
    fpath : str
        Input YAML file path.

    Returns
    -------
    dict
        Attributes read from the YAML file.
    """
    with open(fpath, "r") as f:
        attrs = yaml.safe_load(f)
    return attrs


#### --------------------------------------------------------------------------.
#### Metadata Archive Missing Information


def check_metadata_geolocation(metadata) -> None:
    """Identify metadata with missing or wrong geolocation."""
    # Get longitude, latitude and platform type
    longitude = metadata.get("longitude")
    latitude = metadata.get("latitude")
    platform_type = metadata.get("platform_type")
    # Check type validity
    if isinstance(longitude, str):
        raise TypeError("longitude is not defined as numeric.")
    if isinstance(latitude, str):
        raise TypeError("latitude is not defined as numeric.")
    # Check is not none
    if isinstance(longitude, type(None)) or isinstance(latitude, type(None)):
        raise ValueError("Unspecified longitude and latitude coordinates.")
    else:
        # Check value validity
        # - If mobile platform
        if platform_type == "mobile":
            if longitude != -9999 or latitude != -9999:
                raise ValueError(
                    "For mobile platform_type, specify latitude and longitude -9999"
                )
        # - If fixed platform
        else:
            if longitude == -9999 or latitude == -9999:
                raise ValueError("Missing lat lon coordinates (-9999).")
            elif longitude > 180 or longitude < -180:
                raise ValueError("Unvalid longitude (outside [-180, 180])")
            elif latitude > 90 or latitude < -90:
                raise ValueError("Unvalid latitude (outside [-90, 90])")
            else:
                pass
    return None


def identify_missing_metadata_coords(metadata_fpaths: str) -> None:
    """Identify missing coordinates.

    Parameters
    ----------
    metadata_fpaths : str
        Input YAML file path.

    Raises
    ------
    TypeError
        Error if latitude or longitude coordinates are not present or are wrongly formatted.

    """
    for fpath in metadata_fpaths:
        metadata = read_yaml(fpath)
        check_metadata_geolocation(metadata)
    return None


def identify_empty_metadata_keys(metadata_fpaths: list, keys: Union[str, list]) -> None:
    """Identify empty metadata keys.

    Parameters
    ----------
    metadata_fpaths : str
        Input YAML file path.
    keys : Union[str,list]
        Attributes to verify the presence.
    """

    if isinstance(keys, str):
        keys = [keys]

    for fpath in metadata_fpaths:
        for key in keys:
            metadata = read_yaml(fpath)
            if len(str(metadata.get(key, ""))) == 0:  # ensure is string to avoid error
                print(f"Empty {key} at: ", fpath)
    return None


def get_archive_metadata_key_value(disdrodb_dir, key, return_tuple=True):
    """Return the values of a metadata key for all the archive."""
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    list_info = []
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")
        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        value = metadata[key]
        info = (data_source, campaign_name, station_name, value)
        list_info.append(info)
    if not return_tuple:
        list_info = [info[3] for info in list_info]
    return list_info


#### --------------------------------------------------------------------------.
#### Metadata Archive Checks
def check_archive_metadata_keys(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            _check_metadata_keys(metadata)
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_campaign_name(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            _check_metadata_campaign_name(metadata, expected_name=campaign_name)
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_data_source(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            _check_metadata_data_source(metadata, expected_name=data_source)
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_sensor_name(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            _check_metadata_sensor_name(metadata)
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_station_name(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            _check_metadata_station_name(metadata, expected_name=station_name)
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_reader(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            _check_metadata_reader(metadata)
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_compliance(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")
        try:
            check_metadata_compliance(
                disdrodb_dir=disdrodb_dir,
                data_source=data_source,
                campaign_name=campaign_name,
                station_name=station_name,
            )
        except Exception as e:
            print(f"Error for {data_source} {campaign_name} {station_name}.")
            print(f"The error is: {e}.")


def check_archive_metadata_geolocation(disdrodb_dir):
    list_metadata_paths = get_metadata_list(disdrodb_dir)
    for fpath in list_metadata_paths:
        disdrodb_dir = get_disdrodb_dir(fpath)
        data_source = get_data_source(fpath)
        campaign_name = get_campaign_name(fpath)
        station_name = os.path.basename(fpath).replace(".yml", "")

        metadata = read_station_metadata(
            disdrodb_dir=disdrodb_dir,
            product_level="RAW",
            data_source=data_source,
            campaign_name=campaign_name,
            station_name=station_name,
        )
        try:
            check_metadata_geolocation(metadata)
        except Exception as e:
            print(
                f"Missing information for {data_source} {campaign_name} {station_name}."
            )
            print(f"The error is: {e}.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import numpy as np
from trollsift import Parser

####---------------------------------------------------------------------------
########################
#### FNAME PATTERNS ####
########################
DISDRODB_FNAME_PATTERN = "{product_level:s}.{campaign_name:s}.{station_name:s}.s{start_time:%Y%m%d%H%M%S}.e{end_time:%Y%m%d%H%M%S}.{version:s}.{data_format:s}"

####---------------------------------------------------------------------------.
##########################
#### Filename parsers ####
##########################
fname = "L0A.LOCARNO_2018.60.s20180625004331.e20180711010000.dev.parquet"


def _parse_fname(fname):
    """Parse the filename with trollsift."""
    # Retrieve information from filename
    p = Parser(DISDRODB_FNAME_PATTERN)
    info_dict = p.parse(fname)
    return info_dict


def _get_info_from_filename(fname):
    """Retrieve file information dictionary from filename."""
    # Try to parse the filename
    try:
        info_dict = _parse_fname(fname)
    except ValueError:
        raise ValueError(f"{fname} can not be parsed. Report the issue.")
    # Return info dictionary
    return info_dict


def get_info_from_filepath(fpath):
    """Retrieve file information dictionary from filepath."""
    if not isinstance(fpath, str):
        raise TypeError("'fpath' must be a string.")
    fname = os.path.basename(fpath)
    return _get_info_from_filename(fname)


def get_key_from_filepath(fpath, key):
    """Extract specific key information from a list of filepaths."""
    value = get_info_from_filepath(fpath)[key]
    return value


def get_key_from_filepaths(fpaths, key):
    """Extract specific key information from a list of filepaths."""
    if isinstance(fpaths, str):
        fpaths = [fpaths]
    return [get_key_from_filepath(fpath, key=key) for fpath in fpaths]


####--------------------------------------------------------------------------.
####################################
#### DISDRODB File Informations ####
####################################


def _get_version_from_filepath(filepath, integer=False):
    version = get_key_from_filepath(filepath, key="version")
    if integer:
        version = int(re.findall("\d+", version)[0])
    return version


def get_version_from_filepaths(filepaths, integer=False):
    if isinstance(filepaths, str):
        filepaths = [filepaths]
    list_version = [
        _get_version_from_filepath(fpath, integer=integer) for fpath in filepaths
    ]
    return list_version


def get_campaign_name_from_filepaths(filepaths):
    list_id = get_key_from_filepaths(filepaths, key="campaign_name")
    return list_id


def get_station_name_from_filepaths(filepaths):
    list_id = get_key_from_filepaths(filepaths, key="station_name")
    return list_id


def get_product_level_from_filepaths(filepaths):
    list_id = get_key_from_filepaths(filepaths, key="product_level")
    return list_id


def get_start_time_from_filepaths(filepaths):
    list_start_time = get_key_from_filepaths(filepaths, key="start_time")
    return list_start_time


def get_end_time_from_filepaths(filepaths):
    list_end_time = get_key_from_filepaths(filepaths, key="end_time")
    return list_end_time


def get_start_end_time_from_filepaths(filepaths):
    list_start_time = get_key_from_filepaths(filepaths, key="start_time")
    list_end_time = get_key_from_filepaths(filepaths, key="end_time")
    return np.array(list_start_time), np.array(list_end_time)


####--------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
"""Reader for the EROSION campaign in Denmark."""
from disdrodb.l0 import run_l0b_from_nc
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    # Define dictionary mapping dataset variables to select and rename
    dict_names = {
        ### Dimensions
        "particle_diameter": "diameter_bin_center",
        "particle_fall_speed": "velocity_bin_center",
        ### Variables
        "rain_rate": "rainfall_rate_32bit",
        "rain_amount": "rainfall_accumulated_32bit",
        "SYNOP_4680": "weather_code_synop_4680",
        "SYNOP_4677": "weather_code_synop_4677",
        "reflectivity": "reflectivity_32bit",
        "visibility": "mor_visibility",
        "particles": "number_particles",
        "laser_amplitude": "laser_amplitude",
        "sample_interval": "sample_interval",
        "power_voltage": "sensor_battery_voltage",
        "temp_sensor": "sensor_temperature",
        "heat_status": "sensor_heating_current",
        "status": "sensor_status",
        "rain_kinetic_energy": "rain_kinetic_energy",
        "number_of_particles_classified": "raw_drop_number",
    }

    # Define dataset sanitizer
    def ds_sanitizer_fun(ds):
        pass

        # Return dataset
        return ds

    # Define glob pattern to search data files in <raw_dir>/data/<station_name>
    glob_patterns = "*.nc"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0b_from_nc(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader
        glob_patterns=glob_patterns,
        dict_names=dict_names,
        ds_sanitizer_fun=ds_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )


# -----------------------------------------------------------------.



# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
##################################################
## Wrapper to concat L0B files by command lines ##
##################################################
import sys
import click
from disdrodb.l0.l0_processing import (
    click_l0_station_arguments,
    click_l0b_concat_options,
)

sys.tracebacklimit = 0  # avoid full traceback error if occur


@click.command()
@click_l0_station_arguments
@click_l0b_concat_options
def run_disdrodb_l0b_concat_station(
    # Station arguments
    disdrodb_dir,
    data_source,
    campaign_name,
    station_name,
    # L0B concat options
    remove_l0b=False,
    verbose=True,
):
    """Concatenation all L0B files of a specific DISDRODB station into a single netCDF.

    Parameters
    ----------
    disdrodb_dir : str
        Base directory of DISDRODB
        Format: <...>/DISDRODB
    data_source : str
        Institution name (when campaign data spans more than 1 country),
        or country (when all campaigns (or sensor networks) are inside a given country).
        Must be UPPER CASE.
    campaign_name : str
        Campaign name. Must be UPPER CASE.
    station_name : str
        Station name
    remove_l0b : bool
        If true, remove all source L0B files once L0B concatenation is terminated.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    """
    from disdrodb.api.io import _get_disdrodb_directory

    from disdrodb.l0.L0B_concat import _concatenate_netcdf_files

    # Retrieve processed_dir
    processed_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level="L0B",
        data_source=data_source,
        campaign_name=campaign_name,
        check_exist=True,
    )

    # Run concatenation
    _concatenate_netcdf_files(
        processed_dir=processed_dir,
        station_name=station_name,
        remove=remove_l0b,
        verbose=verbose,
    )

    return None


if __name__ == "__main__":
    run_disdrodb_l0b_concat_station()



# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
##################################################
## Wrapper to concat L0B files by command lines ##
##################################################
import sys
import click
from disdrodb.utils.scripts import parse_arg_to_list
from disdrodb.l0.l0_processing import (
    click_l0_stations_options,
    click_l0b_concat_options,
)

sys.tracebacklimit = 0  # avoid full traceback error if occur


@click.command()
@click.argument("disdrodb_dir", metavar="<DISDRODB base directory>")
@click_l0_stations_options
@click_l0b_concat_options
def run_disdrodb_l0b_concat(
    disdrodb_dir,
    data_sources,
    campaign_names,
    station_names,
    remove_l0b=False,
    verbose=True,
):
    """Run the L0B concatenation of available DISDRODB stations.

    This function enable to launch the processing of many DISDRODB stations with a single command.
    From the list of all available DISDRODB stations, it runs the processing of the
    stations matching the provided data_sources, campaign_names and station_names.

    Parameters
    ----------
    disdrodb_dir : str
        Base directory of DISDRODB
        Format: <...>/DISDRODB
    data_sources : str
        Name of data source(s) to process.
        The name(s) must be UPPER CASE.
        If campaign_names and station are not specified, process all stations.
        To specify multiple data sources, write i.e.: --data_sources 'GPM EPFL NCAR'
    campaign_names : str
        Name of the campaign(s) to process.
        The name(s) must be UPPER CASE.
        To specify multiple campaigns, write i.e.: --campaign_names 'IPEX IMPACTS'
    station_names : str
        Station names.
        To specify multiple stations, write i.e.: --station_names 'station1 station2'
    remove_l0b : bool
        If true, remove all source L0B files once L0B concatenation is terminated.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    """
    from disdrodb.l0.L0B_concat import run_disdrodb_l0b_concat

    # Parse data_sources, campaign_names and station arguments
    data_sources = parse_arg_to_list(data_sources)
    campaign_names = parse_arg_to_list(campaign_names)
    station_names = parse_arg_to_list(station_names)

    # Run concatenation
    run_disdrodb_l0b_concat(
        disdrodb_dir=disdrodb_dir,
        data_sources=data_sources,
        campaign_names=campaign_names,
        station_names=station_names,
        remove=remove,
        verbose=verbose,
    )


if __name__ == "__main__":
    run_disdrodb_l0b_concat()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
"""Reader for CHONGQING campaign."""
from disdrodb.l0 import run_l0a
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    ##------------------------------------------------------------------------.
    #### - Define column names
    column_names = ["TO_SPLIT"]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    reader_kwargs["delimiter"] = ","
    # - No header
    # reader_kwargs["header"] = None
    # - Define encoding
    reader_kwargs["encoding"] = "latin-1"
    # - Avoid first column to become df index
    reader_kwargs["index_col"] = False
    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"
    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"
    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"
    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error", "-.-", "C" * 32]

    ##------------------------------------------------------------------------.
    #### - Define dataframe sanitizer function for L0 processing
    def df_sanitizer_fun(df):
        # - Import pandas
        import numpy as np
        import pandas as pd

        # - Drop unvalid rows
        # --> C*32 that is nan
        df = df.dropna()

        # - Check if there are valid data
        if len(df) == 0:
            raise ValueError("No data available")

        # - Retrieve timesteps
        df_time = df["TO_SPLIT"].iloc[0::33]
        df_time = pd.to_datetime(df_time, format="%Y.%m.%d;%H:%M", errors="coerce")
        df_time = df_time.rename("time")
        df_time = df_time.reset_index(drop=True)
        # - Retrieve data
        idx = np.ones(len(df)).astype(bool)
        idx[0::33] = False
        df_data = df[idx]

        # - Check consistency (no missing rows)
        n_expected_data_rows = int(len(df_time) * 32)
        if len(df_data) != n_expected_data_rows:
            raise ValueError("Not same amount of timesteps and data.")

        # - Replace heterogenous number of spaces with a single space
        df_data["TO_SPLIT"] = (
            df_data["TO_SPLIT"].str.replace(r" +", " ", regex=True).str.strip(" ")
        )

        # - Retrieve arrays
        arr = df_data["TO_SPLIT"].str.split(" ", expand=True).to_numpy()
        # - Flatten by row and then reshape to n_timesteps x 1024
        arr = arr.flatten(order="C").reshape(len(df_time), 1024)
        # - Then concat all the 1024 bins
        df_arr = pd.DataFrame(arr, dtype="str")
        df_raw_drop_number = df_arr.agg(",".join, axis=1)
        df_raw_drop_number = df_raw_drop_number.rename("raw_drop_number")

        # - Create dataframe
        df = pd.concat([df_time, df_raw_drop_number], axis=1)

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in <raw_dir>/data/<station_name>
    glob_patterns = "*.txt*"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0a(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader for L0A processing
        glob_patterns=glob_patterns,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

import os
import glob
import yaml
import numpy as np


def _get_disdrodb_directory(
    disdrodb_dir,
    product_level,
    data_source="",
    campaign_name="",
    check_exist=True,
):
    from disdrodb.l0.io import _check_directory_exist

    # Get directory
    if product_level.upper() == "RAW":
        dir_path = os.path.join(disdrodb_dir, "Raw", data_source, campaign_name)
    else:
        dir_path = os.path.join(disdrodb_dir, "Processed", data_source, campaign_name)
    if check_exist:
        _check_directory_exist(dir_path)
    return dir_path


def _get_list_stations_dirs(product_level, campaign_dir):
    # Get directory where data are stored
    # - Raw: <campaign>/data/<...>
    # - Processed: <campaign>/L0A/L0B>
    if product_level.upper() == "RAW":
        data_path = os.path.join(campaign_dir, "data")
    else:
        data_path = os.path.join(campaign_dir, product_level)
    # Get list of directories (stations)
    list_stations = os.listdir(data_path)
    list_stations_dir = [
        os.path.join(data_path, station_name) for station_name in list_stations
    ]
    return list_stations_dir


def _get_list_stations_with_data(product_level, campaign_dir):
    """Get the list of stations with data inside."""
    # Get stations directory
    list_stations_dir = _get_list_stations_dirs(
        product_level=product_level, campaign_dir=campaign_dir
    )
    # Count number of files within directory
    list_nfiles_per_station = [
        len(glob.glob(os.path.join(path, "*"))) for path in list_stations_dir
    ]
    # Keep only stations with at least one file
    list_stations = [
        os.path.basename(path)
        for n, path in zip(list_nfiles_per_station, list_stations_dir)
        if n >= 1
    ]
    return list_stations


def _get_list_stations_with_metadata(campaign_dir):
    # Get directory where metadata are stored
    metadata_path = os.path.join(campaign_dir, "metadata")
    # List metadata files
    list_metadata_files = glob.glob(os.path.join(metadata_path, "*.yml"))
    # Return stations with metadata
    list_stations = [
        os.path.basename(fpath).replace(".yml", "") for fpath in list_metadata_files
    ]
    return list_stations


def _get_campaign_stations(disdrodb_dir, product_level, data_source, campaign_name):
    # Get campaign directory
    campaign_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_source=data_source,
        campaign_name=campaign_name,
    )

    # Get list of stations with data and metadata
    list_stations_data = _get_list_stations_with_data(
        product_level=product_level, campaign_dir=campaign_dir
    )
    list_stations_metadata = _get_list_stations_with_metadata(campaign_dir)
    # Get list of stations with both data and metadata
    list_stations = list(set(list_stations_data).intersection(list_stations_metadata))

    # Return all available stations for a give campaign
    return list_stations


def _get_campaigns_stations(disdrodb_dir, product_level, data_source, campaign_names):
    if isinstance(campaign_names, str):
        campaign_names = [campaign_names]
    list_available_stations = []
    for campaign_name in campaign_names:
        # Get list of available stations
        list_stations = _get_campaign_stations(
            disdrodb_dir=disdrodb_dir,
            product_level=product_level,
            data_source=data_source,
            campaign_name=campaign_name,
        )
        for station_name in list_stations:
            list_available_stations.append((data_source, campaign_name, station_name))

    # Return all available stations for the asked campaigns (and specific data source)
    return list_available_stations


def _get_data_source_stations(disdrodb_dir, product_level, data_source):
    """Return list of available stations for a specific data source.

    Returns a tuple (<data_source>, <campaign_name>, <station_name>)
    """
    # Get data source directory
    data_source_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_source=data_source,
        campaign_name="",
    )
    # Get possible campaign list
    campaign_names = os.listdir(data_source_dir)

    # For each campaign, retrieve available stations
    list_available_stations = _get_campaigns_stations(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_source=data_source,
        campaign_names=campaign_names,
    )

    # Return all available stations for a specific data source
    return list_available_stations


def _get_data_sources_stations(disdrodb_dir, product_level, data_sources):
    if isinstance(data_sources, str):
        data_sources = [data_sources]
    list_available_stations = []
    for data_source in data_sources:
        list_available = _get_data_source_stations(
            disdrodb_dir=disdrodb_dir,
            product_level=product_level,
            data_source=data_source,
        )
        for station_name in list_available:
            list_available_stations.append(station_name)

    # Return all available stations
    return list_available_stations


def _get_stations(disdrodb_dir, product_level):
    # Get raw or processed directory
    level_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_source="",
        campaign_name="",
    )
    # Get possible data sources
    data_sources = os.listdir(level_dir)

    # For each data_source, retrieve available stations
    list_available_stations = _get_data_sources_stations(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_sources=data_sources,
    )
    # Return all available stations
    return list_available_stations


def _get_metadata_fpath(
    disdrodb_dir, product_level, data_source, campaign_name, station_name
):
    """Get metadata file path a given station."""
    campaign_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_source=data_source,
        campaign_name=campaign_name,
    )
    metadata_fpath = os.path.join(campaign_dir, "metadata", station_name + ".yml")
    return metadata_fpath


def get_metadata_dict(
    disdrodb_dir, product_level, data_source, campaign_name, station_name
):
    """Get metadata of a given station."""
    metadata_fpath = _get_metadata_fpath(
        disdrodb_dir, product_level, data_source, campaign_name, station_name
    )
    with open(metadata_fpath, "r") as f:
        metadata_dict = yaml.safe_load(f)
    return metadata_dict


####---------------------------------------------------------------------------.
#### I/O CHECKS
def check_product_level(product_level):
    """Check DISDRODB product level."""
    if not isinstance(product_level, str):
        raise TypeError("`product_level` must be a string.")
    valid_product_levels = ["RAW", "L0A", "L0B"]
    if product_level.upper() not in valid_product_levels:
        raise ValueError(f"Valid `product_levels` are {valid_product_levels}.")
    return product_level


def check_data_sources(disdrodb_dir, product_level, data_sources):
    """Check DISDRODB data source.

    It checks only if the directory exist.
    """
    # If data_sources is None, return None
    if isinstance(data_sources, type(None)):
        return data_sources
    # Ensure is a list
    if isinstance(data_sources, str):
        data_sources = [data_sources]
    # Remove duplicates
    data_sources = np.unique(np.array(data_sources))
    # Get directory
    dir_path = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir, product_level=product_level
    )
    # Get data sources directory
    list_dir = os.listdir(dir_path)
    # Check if there are unvalid data_sources
    idx_unvalid = np.where(np.isin(data_sources, list_dir, invert=True))[0]
    if len(idx_unvalid) > 0:
        unvalid_data_sources = data_sources[idx_unvalid].tolist()
        raise ValueError(f"These data sources are unvalid: {unvalid_data_sources}.")
    # Return data_sources list
    data_sources = data_sources.tolist()
    return data_sources
    """Check DISDRODB data_sources are valid.
    
    It checks only if the directory exist within the product level.
    """
    # If data_sources is None, return None
    if isinstance(data_sources, type(None)):
        return data_sources
    # Ensure is a list
    if isinstance(data_sources, str):
        data_sources = [data_sources]
    # Remove duplicates
    data_sources = np.unique(np.array(data_sources))
    # Get product level directory path
    dir_path = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir, product_level=product_level
    )
    # Get data sources directory
    list_data_sources = os.listdir(dir_path)
    # Check if there are unvalid data_sources
    idx_unvalid = np.where(np.isin(data_sources, list_data_sources, invert=True))[0]
    if len(idx_unvalid) > 0:
        unvalid_data_sources = data_sources[idx_unvalid].tolist()
        raise ValueError(f"These data sources are unvalid: {unvalid_data_sources}.")
    # Return data_sources list
    data_sources = data_sources.tolist()
    return data_sources


def _check_campaign_names(disdrodb_dir, product_level, campaign_names):
    """Check DISDRODB campaign_names are valid.

    It checks only if the directory exist within the product level.
    """
    # If campaign_names is None, return None
    if isinstance(campaign_names, type(None)):
        return campaign_names
    # Ensure is a list
    if isinstance(campaign_names, str):
        campaign_names = [campaign_names]
    # Remove duplicates
    campaign_names = np.unique(np.array(campaign_names))
    # Get product level directory path
    dir_path = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir, product_level=product_level
    )
    # Get campaigns directory path
    list_campaigns_path = glob.glob(os.path.join(dir_path, "*", "*"))
    # Get campaigns names
    list_campaign_names = [os.path.basename(path) for path in list_campaigns_path]
    # Remove duplicates
    list_campaign_names = np.unique(list_campaign_names)
    # Check if there are unvalid campaign_names
    idx_unvalid = np.where(np.isin(campaign_names, list_campaign_names, invert=True))[0]
    if len(idx_unvalid) > 0:
        unvalid_campaign_names = campaign_names[idx_unvalid].tolist()
        raise ValueError(f"These campaign names are unvalid: {unvalid_campaign_names}.")
    # Return campaign_names list
    campaign_names = campaign_names.tolist()
    return campaign_names


####---------------------------------------------------------------------------.
#### DISDRODB I/O INTERFACE


def available_data_sources(disdrodb_dir, product_level):
    """Return data sources for which stations data are available."""
    product_level = check_product_level(product_level)
    # Get available stations
    list_available_stations = _get_stations(
        disdrodb_dir=disdrodb_dir, product_level=product_level
    )
    data_sources = [info[0] for info in list_available_stations]
    data_sources = np.unique(data_sources).tolist()
    return data_sources


def available_campaigns(
    disdrodb_dir, product_level, data_sources=None, return_tuple=True
):
    """Return campaigns for which stations data are available."""
    # Checks
    product_level = check_product_level(product_level)
    data_sources = check_data_sources(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_sources=data_sources,
    )
    # Get available stations
    if data_sources is None:
        list_available_stations = _get_stations(
            disdrodb_dir=disdrodb_dir, product_level=product_level
        )
    else:
        list_available_stations = _get_data_sources_stations(
            disdrodb_dir=disdrodb_dir,
            product_level=product_level,
            data_sources=data_sources,
        )
    if not return_tuple:
        campaigns = [info[1] for info in list_available_stations]
        campaigns = np.unique(campaigns).tolist()
        return campaigns
    else:
        data_source_campaigns = [(info[0], info[1]) for info in list_available_stations]
        data_source_campaigns = list(set(data_source_campaigns))
        return data_source_campaigns


def available_stations(
    disdrodb_dir,
    product_level,
    data_sources=None,
    campaign_names=None,
    return_tuple=True,
):
    """Return stations for which data are available."""
    # Checks
    product_level = check_product_level(product_level)
    data_sources = check_data_sources(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        data_sources=data_sources,
    )
    campaign_names = _check_campaign_names(
        disdrodb_dir=disdrodb_dir,
        product_level=product_level,
        campaign_names=campaign_names,
    )
    # Format arguments to list
    if isinstance(data_sources, str):
        data_sources = [data_sources]
    if isinstance(campaign_names, str):
        campaign_names = [campaign_names]

    # If data_source is None, first retrieve all stations
    if data_sources is None:
        list_info = _get_stations(
            disdrodb_dir=disdrodb_dir, product_level=product_level
        )
    # Otherwise retrieve all stations for the specified data sources
    else:
        list_info = _get_data_sources_stations(
            disdrodb_dir=disdrodb_dir,
            data_sources=data_sources,
            product_level=product_level,
        )

    # Then, if campaign_name is not None, subset by campaign_name
    if campaign_names is not None:
        list_info = [info for info in list_info if info[1] in campaign_names]

    if return_tuple:
        return list_info
    else:
        # TODO: ENSURE THAT NO DUPLICATED STATION NAMES ?
        list_stations = [info[2] for info in list_info]
        return list_stations



# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import sys
import click
from disdrodb.l0.l0_processing import (
    click_l0_processing_options,
    click_l0_station_arguments,
)

sys.tracebacklimit = 0  # avoid full traceback error if occur

# -------------------------------------------------------------------------.
# Click Command Line Interface decorator


@click.command()
@click_l0_station_arguments
@click_l0_processing_options
def run_disdrodb_l0a_station(
    # Station arguments
    disdrodb_dir,
    data_source,
    campaign_name,
    station_name,
    # Processing options
    force: bool = False,
    verbose: bool = False,
    parallel: bool = True,
    debugging_mode: bool = False,
):
    """Run the L0A processing of a specific DISDRODB station from the terminal.

    Note: this function is used also for processing raw netCDF into L0B format.

    Parameters
    ----------
    disdrodb_dir : str
        Base directory of DISDRODB
        Format: <...>/DISDRODB
    data_source : str
        Institution name (when campaign data spans more than 1 country),
        or country (when all campaigns (or sensor networks) are inside a given country).
        Must be UPPER CASE.
    campaign_name : str
        Campaign name. Must be UPPER CASE.
    station_name : str
        Station name
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is True.
    parallel : bool
        If True, the files are processed simultanously in multiple processes.
        Each process will use a single thread.
        By default, the number of process is defined with os.cpu_count().
        However, you can customize it by typing: DASK_NUM_WORKERS=4 run_disdrodb_l0a_station
        If False, the files are processed sequentially in a single process.
        If False, multi-threading is automatically exploited to speed up I/0 tasks.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        It processes just the first 3 raw data files.
        The default is False.
    """
    import os
    import dask
    from dask.distributed import Client, LocalCluster
    from disdrodb.api.io import _get_disdrodb_directory
    from disdrodb.l0.l0_reader import get_station_reader

    # -------------------------------------------------------------------------.
    # If parallel=True, set the dask environment
    if parallel:
        # Set HDF5_USE_FILE_LOCKING to avoid going stuck with HDF
        os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
        # Retrieve the number of process to run
        available_workers = os.cpu_count() - 2  # if not set, all CPUs
        num_workers = dask.config.get("num_workers", available_workers)
        # Create dask.distributed local cluster
        cluster = LocalCluster(
            n_workers=num_workers,
            threads_per_worker=1,
            processes=True,
            # memory_limit='8GB',
            # silence_logs=False,
        )
        Client(cluster)
    # -------------------------------------------------------------------------.
    # Get reader
    reader = get_station_reader(
        disdrodb_dir=disdrodb_dir,
        data_source=data_source,
        campaign_name=campaign_name,
        station_name=station_name,
    )
    # Define raw_dir and process_dir
    raw_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level="RAW",
        data_source=data_source,
        campaign_name=campaign_name,
    )
    processed_dir = _get_disdrodb_directory(
        disdrodb_dir=disdrodb_dir,
        product_level="L0A",
        data_source=data_source,
        campaign_name=campaign_name,
        check_exist=False,
    )

    # Run L0A processing
    # --> The reader call the run_l0a within the custom defined reader function
    # --> For the special case of raw netCDF data, it calls the run_l0b_from_nc function
    reader(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Processing options
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        parallel=parallel,
    )
    # -------------------------------------------------------------------------.
    # Close the cluster
    if parallel:
        cluster.close()
    return None


if __name__ == "__main__":
    run_disdrodb_l0a_station()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import logging
import numpy as np
import pandas as pd
from disdrodb.l0.standards import (
    get_l0a_dtype,
    get_valid_values_dict,
    get_data_range_dict,
)

# Logger
from disdrodb.utils.logger import (
    log_info,
    log_error,
)

logger = logging.getLogger(__name__)


def _check_valid_range(df, dict_data_range, verbose=False):
    """Check valid value range of dataframe columns.

    It assumes the dict_data_range values are list [min_val, max_val]
    """
    list_wrong_columns = []
    for column in df.columns:
        if column in list(dict_data_range.keys()):
            # If nan occurs, assume it as valid values
            idx_nan = np.isnan(df[column])
            idx_valid = df[column].between(*dict_data_range[column])
            idx_valid = np.logical_or(idx_valid, idx_nan)
            if not idx_valid.all():
                list_wrong_columns.append(column)

    if len(list_wrong_columns) > 0:
        msg = (
            f"Columns {list_wrong_columns} has values outside the expected data range."
        )
        log_error(logger=logger, msg=msg, verbose=False)
        raise ValueError(msg)


def _check_valid_values(df, dict_valid_values, verbose=False):
    """Check valid values of dataframe columns.

    It assumes the dict_valid_values values are list [...].
    """
    list_msg = []
    list_wrong_columns = []
    for column in df.columns:
        if column in list(dict_valid_values.keys()):
            valid_values = dict_valid_values[column]
            # If nan occurs, assume it as valid values
            idx_nan = np.isnan(df[column])
            idx_valid = df[column].isin(valid_values)
            idx_valid = np.logical_or(idx_valid, idx_nan)
            if not idx_valid.all():
                list_wrong_columns.append(column)
                msg = f"The column {column} has values different from {valid_values}."
                list_msg.append(msg)

    if len(list_wrong_columns) > 0:
        msg = "\n".join(list_msg)
        log_error(logger=logger, msg=msg, verbose=False)
        raise ValueError(f"Columns {list_wrong_columns} have invalid values.")


def _check_raw_fields_available(
    df: pd.DataFrame, sensor_name: str, verbose: bool = False
) -> None:
    """Check the presence of the raw spectrum data according to the type of sensor.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe
    sensor_name : str
        Name of the sensor.

    Raises
    ------
    ValueError
        Error if the raw_drop_number field is missing.
    """
    from disdrodb.l0.standards import get_raw_array_nvalues

    # Retrieve raw arrays that could be available (based on sensor_name)
    n_bins_dict = get_raw_array_nvalues(sensor_name=sensor_name)
    raw_vars = np.array(list(n_bins_dict.keys()))

    # Check that raw_drop_number is present
    if "raw_drop_number" not in df.columns:
        msg = "The 'raw_drop_number' column is not present in the dataframe."
        log_error(logger=logger, msg=msg, verbose=False)
        raise ValueError(msg)

    # Report additional raw arrays that are missing
    missing_vars = raw_vars[np.isin(raw_vars, list(df.columns), invert=True)]
    if len(missing_vars) > 0:
        msg = f"The following raw array variable are missing: {missing_vars}"
        log_info(logger=logger, msg=msg, verbose=verbose)
    return None


def check_sensor_name(sensor_name: str) -> None:
    """Check sensor name.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Raises
    ------
    TypeError
        Error if `sensor_name` is not a string.
    ValueError
        Error if the input sensor name has not been found in the list of available sensors.
    """
    from disdrodb.l0.standards import available_sensor_name

    available_sensor_name = available_sensor_name()
    if not isinstance(sensor_name, str):
        raise TypeError("'sensor_name' must be a string.")
    if sensor_name not in available_sensor_name:
        msg = f"{sensor_name} not valid {sensor_name}. Valid values are {available_sensor_name}."
        logger.error(msg)
        raise ValueError(msg)


def check_l0a_column_names(df: pd.DataFrame, sensor_name: str) -> None:
    """Checks that the dataframe columns respects DISDRODB standards.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    sensor_name : str
        Name of the sensor.

    Raises
    ------
    ValueError
        Error if some columns do not meet the DISDRODB standards or if the 'time' column is missing in the dataframe.

    """

    # Get valid columns
    dtype_dict = get_l0a_dtype(sensor_name)
    valid_columns = list(dtype_dict)
    valid_columns = valid_columns + ["time", "latitude", "longitude"]
    valid_columns = set(valid_columns)
    # Get dataframe column names
    df_columns = list(df.columns)
    df_columns = set(df_columns)
    # --------------------------------------------
    # Check there aren't valid columns
    unvalid_columns = list(df_columns.difference(valid_columns))
    if len(unvalid_columns) > 0:
        msg = (
            f"The following columns do no met the DISDRODB standards: {unvalid_columns}"
        )
        logger.error(msg)
        raise ValueError(msg)
    # --------------------------------------------
    # Check time column is present
    if "time" not in df_columns:
        msg = "The 'time' column is missing in the dataframe."
        log_error(logger=logger, msg=msg, verbose=False)
        raise ValueError(msg)
    # --------------------------------------------
    return None


def check_l0a_standards(
    df: pd.DataFrame, sensor_name: str, verbose: bool = True
) -> None:
    """Checks that a file respects the DISDRODB L0A standards.

    Parameters
    ----------
    df : pd.DataFrame
        L0A dataframe.
    sensor_name : str
        Name of the sensor.
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is True.

    Raises
    ------
    ValueError
        Error if some columns have inconsistent values.

    """
    # -------------------------------------
    # Check data range
    dict_data_range = get_data_range_dict(sensor_name)
    _check_valid_range(df=df, dict_data_range=dict_data_range, verbose=verbose)

    # -------------------------------------
    # Check categorical data values
    dict_valid_values = get_valid_values_dict(sensor_name)
    _check_valid_values(df=df, dict_valid_values=dict_valid_values, verbose=verbose)

    # -------------------------------------
    # Check if raw spectrum and 1D derivate exists
    _check_raw_fields_available(df=df, sensor_name=sensor_name, verbose=verbose)

    # -------------------------------------
    # Check if latitude and longitude are columns of the dataframe
    # - They should be only provided if the instrument is moving !!!!
    if "latitude" in df.columns:
        msg = " - The L0A dataframe has column 'latitude'. "
        "This should be included only if the sensor is moving. "
        "Otherwise, specify the 'latitude' in the metadata !"
        log_info(logger=logger, msg=msg, verbose=verbose)

    if "longitude" in df.columns:
        msg = " - The L0A dataframe has column 'longitude'. "
        "This should be included only if the sensor is moving. "
        "Otherwise, specify the 'longitude' in the metadata !"
        log_info(logger=logger, msg=msg, verbose=verbose)

    # -------------------------------------


def check_l0b_standards(x: str) -> None:
    # TODO:
    # - Check for realistic values after having removed the flags !!!!
    pass



from disdrodb.l0.l0_processing import (
    run_l0a,
    run_l0b_from_nc,
    run_disdrodb_l0,
    run_disdrodb_l0_station,
)
from disdrodb.l0.l0_reader import available_readers
from disdrodb.l0.check_metadata import (
    check_archive_metadata_geolocation,
    check_archive_metadata_compliance,
)

__all__ = [
    "run_l0a",
    "run_l0b_from_nc",
    "run_disdrodb_l0",
    "run_disdrodb_l0_station",
    "available_readers",
    "check_archive_metadata_compliance",
    "check_archive_metadata_geolocation",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
from disdrodb.l0 import run_l0b_from_nc
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    # Define dictionary mapping dataset variables to select and rename
    dict_names = {
        ### Dimensions
        "diameter": "diameter_bin_center",
        "fallspeed": "velocity_bin_center",
        ### Variables
        "precipitation_flux": "precipitation_rate",
        "solid_precipitation_flux": "snowfall_rate",
        "precipitation_visibility": "mor_visibility",
        "reflectivity": "reflectivity",
        "present_weather_1m": "weather_code_synop_4680",
        "present_weather_5m": "weather_code_synop_4680_5min",
        "max_hail_diameter": "max_hail_diameter",
        "particle_count": "number_particles",
        # "measurement_quality": "quality_index",
        ### Arrays
        # "drop_size_distribution": "raw_drop_concentration",
        # "drop_velocity_distribution": "raw_drop_average_velocity",
        "size_velocity_distribution": "raw_drop_number",
        ### Variables to discard
        # 'year'
        # 'month'
        # 'day'
        # 'hour'
        # 'minute'
        # 'second'
        # 'day_of_year'
        # 'hydrometeor_type_1m' # Pickering et al., 2019
        # 'hydrometeor_type_5m' # Pickering et al., 2019
    }

    # Define dataset sanitizer
    def ds_sanitizer_fun(ds):
        # Drop coordinates not DISDRODB-compliants
        pass
        # Return dataset
        return ds

    # Define glob pattern to search data files in <raw_dir>/data/<station_name>
    glob_patterns = "*.nc*"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0b_from_nc(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader
        glob_patterns=glob_patterns,
        dict_names=dict_names,
        ds_sanitizer_fun=ds_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )


# -----------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
"""Reader for ARM Thies LPM sensor."""
from disdrodb.l0 import run_l0b_from_nc
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    dict_names = {
        ## Dimensions
        "particle_diameter": "diameter_bin_center",
        "particle_fall_velocity": "velocity_bin_center",
        ### Variables
        # "lat": "latitude",
        # "lon": "longitude",
        # "alt": "altitude",
        # 'base_time': 'base_time',
        # 'time_offset': 'time_offset',
        # 'time_bounds': 'time_bounds',
        # "particle_diameter_bounds",
        # "particle_fall_velocity_bounds"
        "synop_4677_weather_code": "weather_code_synop_4677",
        "metar_4678_weather_code": "weather_code_metar_4678",
        "synop_4680_weather_code": "weather_code_synop_4680",
        "synop_4677_5min_weather_code": "weather_code_synop_4677_5min",
        "metar_4678_5min_weather_code": "weather_code_metar_4678_5min",
        "synop_4680_5min_weather_code": "weather_code_synop_4680_5min",
        "intensity_total_5min": "precipitation_rate_5min",
        "intensity_total": "precipitation_rate",
        "intensity_liquid": "rainfall_rate",
        "intensity_solid": "snowfall_rate",
        "accum_precip": "precipitation_accumulated",
        "maximum_visibility": "mor_visibility",
        "radar_reflectivity": "reflectivity",
        "quality_measurement": "quality_index",
        "max_diameter_hail": "max_hail_diameter",
        "laser_status": "laser_status",
        "static_signal": "static_signal",
        "interior_temperature": "temperature_interior",
        "laser_temperature": "laser_temperature",
        "laser_temperature_analog_status": "laser_temperature_analog_status",
        "laser_temperature_digital_status": "laser_temperature_digital_status",
        "mean_laser_current": "laser_current_average",
        "laser_current_analog_status": "laser_current_analog_status",
        "laser_current_digital_status": "laser_current_digital_status",
        "control_voltage": "control_voltage",
        "optical_control_output": "optical_control_voltage_output",
        "control_output_laser_power_status": "control_output_laser_power_status",
        "voltage_sensor_supply": "sensor_voltage_supply",
        "voltage_sensor_supply_status": "sensor_voltage_supply_status",
        "ambient_temperature": "temperature_ambient",
        "temperature_sensor_status": "temperature_sensor_status",
        "voltage_heating_supply": "current_heating_voltage_supply",
        "voltage_heating_supply_status": "current_heating_voltage_supply_status",
        "pane_heating_laser_head_current": "current_heating_pane_transmitter_head",
        "pane_heating_laser_head_current_status": "current_heating_pane_transmitter_head_status",
        "pane_heating_receiver_head_current": "current_heating_pane_receiver_head",
        "pane_heating_receiver_head_current_status": "current_heating_pane_receiver_head_status",
        "heating_house_current": "current_heating_house",
        "heating_house_current_status": "current_heating_house_status",
        "heating_heads_current": "current_heating_heads",
        "heating_heads_current_status": "current_heating_heads_status",
        "heating_carriers_current": "current_heating_carriers",
        "heating_carriers_current_status": "current_heating_carriers_status",
        "number_particles": "number_particles",
        "number_particles_internal_data": "number_particles_internal_data",
        "number_particles_min_speed": "number_particles_min_speed",
        "number_particles_min_speed_internal_data": "number_particles_min_speed_internal_data",
        "number_particles_max_speed": "number_particles_max_speed",
        "number_particles_max_speed_internal_data": "number_particles_max_speed_internal_data",
        "number_particles_min_diameter": "number_particles_min_diameter",
        "number_particles_min_diameter_internal_data": "number_particles_min_diameter_internal_data",
        "precipitation_spectrum": "raw_drop_number",
        # 'air_temperature',
    }

    # Define dataset sanitizer
    def ds_sanitizer_fun(ds):
        from disdrodb.l0.l0b_processing import replace_custom_nan_flags

        # Replace nan flags with np.nan
        # - ARM use the -9999 flags
        nan_flags_variables = [
            "weather_code_synop_4677_5min",
            "weather_code_synop_4680_5min",
            "weather_code_synop_4680",
            "weather_code_synop_4677",
            "mor_visibility",
            "quality_index",
            "temperature_interior",
            "laser_temperature",
            "laser_current_average",
            "number_particles",
            "number_particles_min_speed",
            "number_particles_max_speed",
            "number_particles_min_diameter",
            "number_particles_max_diameter",
            "raw_drop_number",
        ]
        dict_nan_flags = {var: [-9999] for var in nan_flags_variables}
        ds = replace_custom_nan_flags(ds, dict_nan_flags=dict_nan_flags)

        # Return dataset
        return ds

    # Define glob pattern to search data files in <raw_dir>/data/<station_name>
    glob_patterns = "*.nc"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0b_from_nc(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader
        glob_patterns=glob_patterns,
        dict_names=dict_names,
        ds_sanitizer_fun=ds_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )


# -----------------------------------------------------------------.



# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import sys
import click
from disdrodb.l0.l0_processing import (
    click_l0_processing_options,
    click_l0_station_arguments,
    click_l0_archive_options,
)

sys.tracebacklimit = 0  # avoid full traceback error if occur

# -------------------------------------------------------------------------.
# Click Command Line Interface decorator


@click.command()
@click_l0_station_arguments
@click_l0_processing_options
@click_l0_archive_options
def run_disdrodb_l0_station(
    # Station arguments
    disdrodb_dir,
    data_source,
    campaign_name,
    station_name,
    # L0 archive options
    l0a_processing: bool = True,
    l0b_processing: bool = True,
    l0b_concat: bool = True,
    remove_l0a: bool = False,
    remove_l0b: bool = False,
    # Processing options
    force: bool = False,
    verbose: bool = True,
    parallel: bool = True,
    debugging_mode: bool = False,
):
    """Run the L0 processing of a specific DISDRODB station from the terminal.

    Parameters \n
    ---------- \n

    disdrodb_dir : str \n
        Base directory of DISDRODB \n
        Format: <...>/DISDRODB \n
    data_source : str \n
        Institution name (when campaign data spans more than 1 country), or country (when all campaigns (or sensor networks) are inside a given country).\n
        Must be UPPER CASE.\n
    campaign_name : str \n
        Campaign name. Must be UPPER CASE.\n
    station_name : str \n
        Station name \n
    l0a_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.\n
    l0b_processing : bool \n
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data.\n
      The default is True.\n

    l0b_concat : bool \n
        Whether to concatenate all raw files into a single L0B netCDF file.\n
        If l0b_concat=True, all raw files will be saved into a single L0B netCDF file.\n
        If l0b_concat=False, each raw file will be converted into the corresponding L0B netCDF file.\n
        The default is False.\n
    remove_l0a : bool \n
        Whether to keep the L0A files after having generated the L0B netCDF products.\n
        The default is False.\n
    remove_l0b : bool \n
         Whether to remove the L0B files after having concatenated all L0B netCDF files.\n
         It takes places only if l0b_concat=True\n
         The default is False.\n
    force : bool \n
        If True, overwrite existing data into destination directories.\n
        If False, raise an error if there are already data into destination directories.\n
        The default is False.\n
    verbose : bool \n
        Whether to print detailed processing information into terminal.\n
        The default is True.\n
    parallel : bool \n
        If True, the files are processed simultanously in multiple processes.\n
        Each process will use a single thread to avoid issues with the HDF/netCDF library.\n
        By default, the number of process is defined with os.cpu_count().\n
        However, you can customize it by typing: DASK_NUM_WORKERS=4 run_disdrodb_l0_station\n
        If False, the files are processed sequentially in a single process.\n
        If False, multi-threading is automatically exploited to speed up I/0 tasks.\n
    debugging_mode : bool \n
        If True, it reduces the amount of data to process.\n
        For L0A, it processes just the first 3 raw data files for each station.\n
        For L0B, it processes just the first 100 rows of 3 L0A files for each station.\n
        The default is False.\n
    """
    from disdrodb.l0.l0_processing import run_disdrodb_l0_station

    run_disdrodb_l0_station(
        disdrodb_dir=disdrodb_dir,
        data_source=data_source,
        campaign_name=campaign_name,
        station_name=station_name,
        # L0 archive options
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        l0b_concat=l0b_concat,
        remove_l0a=remove_l0a,
        remove_l0b=remove_l0b,
        # Processing options
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        parallel=parallel,
    )

    return None


if __name__ == "__main__":
    run_disdrodb_l0_station()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
"""Reader for the GID LPM network."""
from disdrodb.l0 import run_l0a
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    ##------------------------------------------------------------------------.
    #### - Define column names
    column_names = ["TO_BE_SPLITTED"]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = "\n"
    # Skip first row as columns names
    reader_kwargs["header"] = None
    # - Avoid first column to become df index
    reader_kwargs["index_col"] = False
    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"
    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"
    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"
    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error", "NA", "-.-"]

    ##------------------------------------------------------------------------.
    #### - Define dataframe sanitizer function for L0 processing
    def df_sanitizer_fun(df):
        # - Import pandas
        import pandas as pd

        # - Split columns
        df = df["TO_BE_SPLITTED"].str.split(";", n=79, expand=True)

        # - Assign column names
        column_names = [
            "start_identifier",
            "sensor_serial_number",
            "software_version",
            "sensor_date",
            "sensor_time",
            "weather_code_synop_4677_5min",
            "weather_code_synop_4680_5min",
            "weather_code_metar_4678_5min",
            "precipitation_rate_5min",
            "weather_code_synop_4677",
            "weather_code_synop_4680",
            "weather_code_metar_4678",
            "precipitation_rate",
            "rainfall_rate",
            "snowfall_rate",
            "precipitation_accumulated",
            "mor_visibility",
            "reflectivity",
            "quality_index",
            "max_hail_diameter",
            "laser_status",
            "static_signal",
            "laser_temperature_analog_status",
            "laser_temperature_digital_status",
            "laser_current_analog_status",
            "laser_current_digital_status",
            "sensor_voltage_supply_status",
            "current_heating_pane_transmitter_head_status",
            "current_heating_pane_receiver_head_status",
            "temperature_sensor_status",
            "current_heating_voltage_supply_status",
            "current_heating_house_status",
            "current_heating_heads_status",
            "current_heating_carriers_status",
            "control_output_laser_power_status",
            "reserve_status",
            "temperature_interior",
            "laser_temperature",
            "laser_current_average",
            "control_voltage",
            "optical_control_voltage_output",
            "sensor_voltage_supply",
            "current_heating_pane_transmitter_head",
            "current_heating_pane_receiver_head",
            "temperature_ambient",
            "current_heating_voltage_supply",
            "current_heating_house",
            "current_heating_heads",
            "current_heating_carriers",
            "number_particles",
            "number_particles_internal_data",
            "number_particles_min_speed",
            "number_particles_min_speed_internal_data",
            "number_particles_max_speed",
            "number_particles_max_speed_internal_data",
            "number_particles_min_diameter",
            "number_particles_min_diameter_internal_data",
            "number_particles_no_hydrometeor",
            "number_particles_no_hydrometeor_internal_data",
            "number_particles_unknown_classification",
            "number_particles_unknown_classification_internal_data",
            "number_particles_class_1",
            "number_particles_class_1_internal_data",
            "number_particles_class_2",
            "number_particles_class_2_internal_data",
            "number_particles_class_3",
            "number_particles_class_3_internal_data",
            "number_particles_class_4",
            "number_particles_class_4_internal_data",
            "number_particles_class_5",
            "number_particles_class_5_internal_data",
            "number_particles_class_6",
            "number_particles_class_6_internal_data",
            "number_particles_class_7",
            "number_particles_class_7_internal_data",
            "number_particles_class_8",
            "number_particles_class_8_internal_data",
            "number_particles_class_9",
            "number_particles_class_9_internal_data",
            "raw_drop_number",
        ]
        df.columns = column_names

        # - Remove checksum at end of raw_drop_number
        df["raw_drop_number"] = df["raw_drop_number"].str.slice(stop=1760)

        # - Define 'time column
        df["time"] = df["sensor_date"].astype(str) + " " + df["sensor_time"].astype(str)

        # - Convert time column to datetime
        df["time"] = pd.to_datetime(
            df["time"], format="%d.%m.%y %H:%M:%S", errors="coerce"
        )

        # - Drop columns not agreeing with DISDRODB L0 standards
        columns_to_drop = [
            "start_identifier",
            "software_version",
            "sensor_serial_number",
            "sensor_date",
            "sensor_time",
        ]
        df = df.drop(columns=columns_to_drop)

        df = df.drop(columns=["sensor_date", "sensor_time"])

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in <raw_dir>/data/<station_name>
    glob_patterns = "*.txt*"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0a(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader for L0A processing
        glob_patterns=glob_patterns,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )



from importlib.metadata import version, PackageNotFoundError

from disdrodb.api.io import (
    available_stations,
    available_campaigns,
    available_data_sources,
)
from disdrodb.api.metadata import read_station_metadata
from disdrodb.l0.standards import available_sensor_name

__all__ = [
    "available_stations",
    "available_campaigns",
    "available_data_sources",
    "available_sensor_name",
    "read_station_metadata",
]

# Get version
try:
    __version__ = version("disdrodb")
except PackageNotFoundError:
    # package is not installed
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
from disdrodb.l0 import run_l0a
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    ##------------------------------------------------------------------------.
    #### - Define column names
    column_names = [
        "logger_time",
        "record_number",
        "battery_voltage",
        "logger_temperature",
        "wind_direction",
        "wind_speed",
        "wind_flag",
        "fast_temperature",
        "slow_temperature",
        "relative_humidity",
        "pressure",
        "compass_direction",
        "gps_time",
        "gps_status",
        "gps_latitude",  # DD.DM, DM = decimal minutes/100, DD = degrees
        "gps_latitude_hemisphere",  # N/S
        "gps_longitude",
        "gps_longitude_hemisphere",  # W/E
        "gps_speed",
        "gps_direction",
        "gps_date",
        "gps_magnetic_version",
        "gps_altitude",
        "relative_wind_direction",
        "dew_point_temperature",
        "rh",
        "disdrometer_data",
    ]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = ","

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define encoding
    reader_kwargs["encoding"] = "ISO-8859-1"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error"]

    # Skip first row as columns names
    reader_kwargs["header"] = None

    ##------------------------------------------------------------------------.
    #### - Define dataframe sanitizer function for L0 processing
    def df_sanitizer_fun(df):
        # - Import pandas
        import pandas as pd

        # - Drop timesteps without disdrometer data
        df = df[~df["disdrometer_data"].isna()]

        # - Retrieve disdrometer data
        df_data = df["disdrometer_data"].str.split(";", expand=True, n=11)
        # - Assign column names
        column_names = [
            "serial_number",
            "rainfall_rate_32bit",
            "rainfall_accumulated_32bit",
            "reflectivity_32bit",
            "sample_interval",
            "laser_amplitude",
            "number_particles",
            "sensor_temperature",
            "sensor_battery_voltage",
            "sensor_time",  # Note: logger_time is currently used !
            "sensor_date",
            "raw_drop_number",
        ]
        df_data.columns = column_names

        # - Retrieve time and coordinates information
        # --> Latitude in degrees_north
        # --> Longitude in degrees_east
        df_time = pd.to_datetime(df["logger_time"], errors="coerce")
        df_lat_sign = (
            df["gps_latitude_hemisphere"].str.replace("N", "1").str.replace("S", "-1")
        )
        df_lon_sign = (
            df["gps_longitude_hemisphere"].str.replace("E", "1").str.replace("W", "-1")
        )
        df_lat_sign = df_lat_sign.astype(float)
        df_lon_sign = df_lon_sign.astype(float)
        df_lon = df["gps_longitude"].astype(float)
        df_lat = df["gps_latitude"].astype(float)
        df_lon = df_lon * df_lon_sign
        df_lat = df_lat * df_lat_sign

        # Create dataframe
        df = df_data
        df["time"] = df_time
        df["latitude"] = df_lat
        df["longitude"] = df_lon

        # - Drop columns not agreeing with DISDRODB L0 standards
        df = df.drop(columns=["serial_number", "sensor_time", "serial_number"])

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in <raw_dir>/data/<station_name>
    glob_patterns = "*.txt"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0a(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader for L0A processing
        glob_patterns=glob_patterns,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

import logging
import numpy as np
import pandas as pd
import xarray as xr
from typing import Tuple
from disdrodb.utils.logger import log_info, log_warning, log_error

logger = logging.getLogger(__name__)


####---------------------------------------------------------------------------.
def _sort_datasets_by_dim(
    list_ds: list, fpaths: str, dim: str = "time"
) -> Tuple[list, list]:
    """Sort a list of xarray.Dataset and corresponding file paths by the starting value of a specified dimension.

    Parameters
    ----------
    fpaths : list
        List of netCDFs file paths.
    list_ds : list
        List of xarray Dataset.
    dim : str, optional
        Dimension name. The default is "time".

    Returns
    -------
    tuple
        Tuple of sorted list of xarray datasets and sorted list of file paths.
    """
    start_values = [ds[dim].values[0] for ds in list_ds]
    sorted_idx = np.argsort(start_values)
    sorted_list_ds = [list_ds[i] for i in sorted_idx]
    sorted_fpaths = [fpaths[i] for i in sorted_idx]
    return sorted_list_ds, sorted_fpaths


def _get_dim_values_index(
    list_ds: list, dim: str
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Get list and dataset indices associated to the dimension values."""
    dim_values = np.concatenate([ds[dim].values for ds in list_ds])
    list_index = np.concatenate(
        [np.ones(len(ds[dim])) * i for i, ds in enumerate(list_ds)]
    )
    list_index = list_index.astype(int)
    ds_index = np.concatenate(
        [np.arange(0, len(ds[dim])) for i, ds in enumerate(list_ds)]
    )
    return dim_values, list_index, ds_index


def _get_non_monotonic_indices_to_remove(dim_values: np.ndarray) -> np.ndarray:
    """Returns the indices that cause a non-monotonic increasing series of values.

    Assume that duplicated values, if present, occurs consecutively !
    """
    diff_dim_values = np.diff(dim_values)
    indices_decreasing = np.where(diff_dim_values.astype(float) <= 0)[0] + 1
    if len(indices_decreasing) == 0:
        return []
    idx_start_decreasing = indices_decreasing[0]
    idx_restart_increase = np.max(
        np.where(dim_values <= dim_values[idx_start_decreasing - 1])[0]
    )
    idx_to_remove = np.arange(idx_start_decreasing, idx_restart_increase + 1)
    return idx_to_remove


def _get_duplicated_indices(x, keep="first"):
    """Return the indices to remove for duplicated values in x such that there is only one value occurence.

    Parameters
    ----------
    x :  np.array
        Array of values.
    keep : str, optional
        The value to keep, either 'first', 'last' or False.
        The default is 'first'.
        ‘first’ : Mark duplicates as True except for the first occurrence.
        ‘last’ : Mark duplicates as True except for the last occurrence.
        False : Mark all duplicates as True.

    Returns
    -------
    np.array
        Array of indices to remove.
    """
    # Check 'keep' argument
    # if not isinstance(keep, str):
    #     raise TypeError("`keep` must be a string. Either first or last.")
    # if not np.isin(keep, ["first", "last"]):
    #     raise ValueError("Invalid value for argument keep. Only 'first' and 'last' are accepted.")
    # # Get

    # x_indices = np.arange(len(x))
    # unique_values, unique_counts = np.unique(x, return_counts=True)
    # duplicated_values = unique_values[unique_counts > 1]

    # duplicated_indices = np.array([], dtype=np.int32)
    # if keep == 'first':
    #     for value in duplicated_values:
    #         indices = np.where(x == value)[0]
    #         duplicated_indices = np.concatenate([duplicated_indices, indices[1:]])
    # elif keep == 'last':
    #     indices = np.where(x == value)[0]
    #     duplicated_indices = np.concatenate([duplicated_indices, indices[:-1]])
    # return duplicated_indices

    # Get duplicate indices
    idx_duplicated = pd.Index(x).duplicated(keep=keep)
    return np.where(idx_duplicated)[0]


def _get_bad_info_dict(
    idx_to_remove: np.ndarray,
    list_index: np.ndarray,
    dim_values: np.ndarray,
    ds_index: np.ndarray,
) -> Tuple[dict, dict]:
    """Return two dictionaries mapping, for each dataset, the bad values and indices to remove.

    Parameters
    ----------
    idx_to_remove : np.ndarray
        Indices to be removed to ensure monotonic dimension.
    list_index : np.ndarray
        Indices corresponding to the file in the `list_ds` parameter.
    ds_index : np.ndarray
        Indices corresponding to the dataset dimension index in the `list_ds` parameter.

    Returns
    -------
    dict
        A dictionary mapping the dimension values to remove for each file.
    dict
        A dictionary mapping the dataset dimension indices to remove for each file.
    """
    list_index_bad = list_index[idx_to_remove]
    ds_index_bad = ds_index[idx_to_remove]
    dim_values_bad = dim_values[idx_to_remove]
    # Retrieve dictionary with the bad values in each dataset
    dict_ds_bad_values = {
        k: dim_values_bad[np.where(list_index_bad == k)[0]]
        for k in np.unique(list_index_bad)
    }
    # Retrieve dictionary with the index with the bad values in each dataset
    dict_ds_bad_idx = {
        k: ds_index_bad[np.where(list_index_bad == k)[0]]
        for k in np.unique(list_index_bad)
    }
    return dict_ds_bad_values, dict_ds_bad_idx


def _remove_dataset_bad_values(list_ds, fpaths, dict_ds_bad_idx, dim):
    """Remove portions of xarray Datasets corresponding to duplicated values.

    Parameters
    ----------
    list_ds : list
        List of xarray Dataset.
    dict_ds_bad_idx : dict
        Dictionary with the dimension indices corresponding to bad values in each xarray Dataset.

    Returns
    -------

    list_ds : list
        List of xarray Dataset without bad values.
    """
    list_index_valid = list(range(len(list_ds)))
    for list_index_bad, bad_idx in dict_ds_bad_idx.items():
        # Get dataset
        ds = list_ds[list_index_bad]
        # If resulting in a empty dataset, drop index from list_index_valid
        if len(bad_idx) == len(list_ds[list_index_bad][dim]):
            list_index_valid.remove(list_index_bad)
        # Remove unvalid indices
        list_ds[list_index_bad] = ds.drop_isel({dim: bad_idx})

    # Keep non-empty datasets
    new_list_ds = [list_ds[idx] for idx in list_index_valid]
    new_fpaths = [fpaths[idx] for idx in list_index_valid]
    return new_list_ds, new_fpaths


def ensure_unique_dimension_values(
    list_ds: list, fpaths: str, dim: str = "time", verbose: bool = False
) -> list:
    """Ensure that a list of xr.Dataset has non duplicated dimension values.

    Parameters
    ----------
    list_ds : list
        List of xarray Dataset.
    fpaths : list
        List of netCDFs file paths.
    dim : str, optional
        Dimension name.
        The default is "time".

    Returns
    -------
    list
        List of xarray Dataset.
    list
        List of netCDFs file paths.
    """
    # Reorder the files and filepaths by the starting dimension value (time)
    sorted_list_ds, sorted_fpaths = _sort_datasets_by_dim(
        list_ds=list_ds, fpaths=fpaths, dim=dim
    )

    # Get the datasets dimension values array (and associated list_ds/xr.Dataset indices)
    dim_values, list_index, ds_index = _get_dim_values_index(list_ds, dim=dim)

    # Get duplicated indices
    idx_duplicated = _get_duplicated_indices(dim_values, keep="first")

    # Remove duplicated indices
    if len(idx_duplicated) > 0:
        # Retrieve dictionary providing bad values and indexes for each dataset
        dict_ds_bad_values, dict_ds_bad_idx = _get_bad_info_dict(
            idx_to_remove=idx_duplicated,
            list_index=list_index,
            dim_values=dim_values,
            ds_index=ds_index,
        )

        # Report for each dataset, the duplicates values occuring
        for list_index_bad, bad_values in dict_ds_bad_values.items():
            # Retrieve dataset filepath
            fpath = fpaths[list_index_bad]
            # If all values inside the file are duplicated, report it
            if len(bad_values) == len(list_ds[list_index_bad][dim]):
                msg = f"{fpath} is excluded from concatenation. All {dim} values are already present in some other file."
                log_warning(logger=logger, msg=msg, verbose=verbose)
            else:
                if np.issubdtype(bad_values.dtype, np.datetime64):
                    bad_values = bad_values.astype("M8[s]")
                msg = f"In {fpath}, dropping {dim} values {bad_values} to avoid duplicated {dim} values."
                log_warning(logger=logger, msg=msg, verbose=verbose)

        # Remove duplicated values
        list_ds, fpaths = _remove_dataset_bad_values(
            list_ds=list_ds, fpaths=fpaths, dict_ds_bad_idx=dict_ds_bad_idx, dim=dim
        )
    return list_ds, fpaths


def ensure_monotonic_dimension(
    list_ds: list, fpaths: str, dim: str = "time", verbose: bool = False
) -> list:
    """Ensure that a list of xr.Dataset has a monotonic increasing (non duplicated) dimension values.

    Parameters
    ----------
    list_ds : list
        List of xarray Dataset.
    fpaths : list
        List of netCDFs file paths.
    dim : str, optional
        Dimension name.
        The default is "time".

    Returns
    -------
    list
        List of xarray Dataset.
    list
        List of netCDFs file paths.
    """
    # Reorder the files and filepaths by the starting dimension value (time)
    # TODO: should maybe also split by non-continuous time ...
    sorted_list_ds, sorted_fpaths = _sort_datasets_by_dim(
        list_ds=list_ds, fpaths=fpaths, dim=dim
    )

    # Get the datasets dimension values array (and associated list_ds/xr.Dataset indices)
    dim_values, list_index, ds_index = _get_dim_values_index(list_ds, dim=dim)

    # Identify the indices to remove to ensure monotonic values
    idx_to_remove = _get_non_monotonic_indices_to_remove(dim_values)

    # Remove indices causing the values to be non-monotonic increasing
    if len(idx_to_remove) > 0:
        # Retrieve dictionary providing bad values and indexes for each dataset
        dict_ds_bad_values, dict_ds_bad_idx = _get_bad_info_dict(
            idx_to_remove=idx_to_remove,
            list_index=list_index,
            dim_values=dim_values,
            ds_index=ds_index,
        )

        # Report for each dataset, the values to be dropped
        for list_index_bad, bad_values in dict_ds_bad_values.items():
            # Retrieve dataset filepath
            fpath = fpaths[list_index_bad]
            # If all values inside the file shoudl be dropped, report it
            if len(bad_values) == len(list_ds[list_index_bad][dim]):
                msg = f"{fpath} is excluded from concatenation. All {dim} values cause the dimension to be non-monotonic."
                log_warning(logger=logger, msg=msg, verbose=verbose)
            else:
                if np.issubdtype(bad_values.dtype, np.datetime64):
                    bad_values = bad_values.astype("M8[s]")
                msg = f"In {fpath}, dropping {dim} values {bad_values} to ensure monotonic {dim} dimension."
                log_warning(logger=logger, msg=msg, verbose=verbose)

        # Remove duplicated values
        list_ds, fpaths = _remove_dataset_bad_values(
            list_ds=list_ds, fpaths=fpaths, dict_ds_bad_idx=dict_ds_bad_idx, dim=dim
        )
        # Iterative check
        list_ds, fpathsa = ensure_monotonic_dimension(
            list_ds=list_ds, fpaths=fpaths, dim=dim
        )

    return list_ds, fpaths


# ds_index = [0,1,2,3,0,1,2,3,4]
# list_index = [0,0,0,0,1, 1, 1,1, 1]
# dim_values = [0,1,5,5,5, 5, 6,7,8]
# list_index = np.array(list_index)
# dim_values = np.array(dim_values)
# ds_index = np.array(ds_index)


####---------------------------------------------------------------------------
def get_list_ds(fpaths: str) -> list:
    """Get list of xarray datasets from file paths.

    Parameters
    ----------
    fpaths : list
        List of netCDFs file paths.

    Returns
    -------
    list
        List of xarray datasets.
    """
    import xarray as xr

    list_ds = []
    for fpath in fpaths:
        # This context manager is required to avoid random HDF locking
        # - cache=True: store data in memory to avoid reading back from disk
        # --> but LRU cache might cause the netCDF to not be closed !
        with xr.open_dataset(fpath, cache=False) as data:
            ds = data.load()
        list_ds.append(ds)
    return list_ds


# def get_list_ds(fpaths: str) -> list:
#     """Get list of xarray datasets from file paths.

#     Parameters
#     ----------
#     fpaths : list
#         List of netCDFs file paths.

#     Returns
#     -------
#     list
#         List of xarray datasets.
#     """
#     # WARNING: READING IN PARALLEL USING MULTIPROCESS CAUSE HDF LOCK ERRORS
#     @dask.delayed
#     def open_dataset_delayed(fpath):
#         import os

#         os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
#         import xarray as xr

#         # This context manager is required to avoid random HDF locking
#         # - cache=True: store data in memory to avoid reading back from disk
#         # --> but LRU cache might cause the netCDF to not be closed !
#         with xr.open_dataset(fpath, cache=False) as data:
#             ds = data.load()
#         return ds

#     list_ds_delayed = []
#     for fpath in fpaths:
#         list_ds_delayed.append(open_dataset_delayed(fpath))
#     list_ds = dask.compute(list_ds_delayed)[0]
#     return list_ds


####---------------------------------------------------------------------------


def _concatenate_datasets(list_ds, dim="time", verbose=False):
    try:
        msg = "Start concatenating with xr.concat."
        log_info(logger=logger, msg=msg, verbose=verbose)

        ds = xr.concat(list_ds, dim="time", coords="minimal", compat="override")

        msg = "Concatenation with xr.concat has been successful."
        log_info(logger=logger, msg=msg, verbose=verbose)

    except Exception as e:
        msg = f"Concatenation with xr.concat failed. Error is {e}."
        log_error(logger=logger, msg=msg, verbose=False)
        raise ValueError(msg)
    return ds


def xr_concat_datasets(fpaths: str, verbose=False) -> xr.Dataset:
    """Concat xr.Dataset in a robust and parallel way.

    1. It checks for time dimension monotonicity

    Parameters
    ----------
    fpaths : list
        List of netCDFs file paths.

    Returns
    -------
    xr.Dataset
        A single xarray dataset.

    Raises
    ------
    ValueError
        Error if the merging/concatenation operations can not be achieved.

    """

    # --------------------------------------.
    # Open xr.Dataset lazily in parallel using dask delayed
    list_ds = get_list_ds(fpaths)

    # --------------------------------------.
    # Ensure time dimension contains no duplicated values
    list_ds, fpaths = ensure_unique_dimension_values(
        list_ds=list_ds, fpaths=fpaths, dim="time", verbose=verbose
    )

    # Ensure time dimension is monotonic increasingly
    list_ds, fpaths = ensure_monotonic_dimension(
        list_ds=list_ds, fpaths=fpaths, dim="time", verbose=verbose
    )

    # --------------------------------------.
    # Concatenate all netCDFs
    ds = _concatenate_datasets(list_ds=list_ds, dim="time", verbose=verbose)

    # --------------------------------------.
    # Return xr.Dataset
    return ds



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from asyncio.log import logger
import os
import time
import logging
import re


def create_l0_logger(
    processed_dir: str, campaign_name: str, verbose: bool = False
) -> logger:
    """Create L0 logger.

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory.
    campaign_name : str
        Campaign name.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.

    Returns
    -------
    logger
        Logger object.
    """
    # Define log name
    logger_name = "LO_" + "reader_" + campaign_name
    # Create logs directory
    logs_dir = os.path.join(processed_dir, "logs")
    os.makedirs(logs_dir, exist_ok=True)

    # Create logger
    _create_logger(logs_dir, logger_name)
    logger = logging.getLogger(campaign_name)

    # logger = _create_logger(logs_dir, logger_name)
    # -------------------------------------------------.
    # Update logger
    msg = "### Script started ###"
    if verbose:
        print("\n  " + msg + "\n")
    logger.info(msg)
    # -------------------------------------------------.
    # Return logger
    return logger


def _create_logger(log_dir, logger_name):
    # Define log file filepath
    logger_fname = f'{logger_name}_{time.strftime("%d-%m-%Y_%H-%M-%S")}.log'
    logger_fpath = os.path.join(log_dir, logger_fname)
    # -------------------------------------------------------------------------.
    # Define logger format
    format_type = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    # Define logger level
    level = logging.DEBUG

    # Define logging
    logging.basicConfig(format=format_type, level=level, filename=logger_fpath)

    # Retrieve logger
    # logger = logging.getLogger(logger_fpath)
    # return logger
    return None


def close_logger(logger: logger) -> None:
    """Close the logger

    Parameters
    ----------
    logger : logger
        Logger object.
    """
    handlers = logger.handlers[:]
    for handler in handlers:
        handler.flush()
        handler.close()
        logger.removeHandler(handler)
    return


def create_file_logger(processed_dir, product_level, station_name, filename, parallel):
    ##------------------------------------------------------------------------.
    # Create logs directory
    logs_dir = os.path.join(processed_dir, "logs", product_level, station_name)
    os.makedirs(logs_dir, exist_ok=True)

    # logger_fname = f'logs_{fname}_{time.strftime("%d-%m-%Y_%H-%M-%S")}.log'
    logger_fname = f"logs_{filename}.log"
    logger_fpath = os.path.join(logs_dir, logger_fname)

    # -------------------------------------------------------------------------.
    # Set logger (TODO: messy with multiprocess)
    if parallel:
        logger = logging.getLogger(filename)  # does not log submodules logs
    else:
        logger = logging.getLogger()  # root logger (messy with multiprocess)

    handler = logging.FileHandler(logger_fpath, mode="w")
    # handler.setLevel(logging.DEBUG)
    format_type = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    handler.setFormatter(logging.Formatter(format_type))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return logger


####---------------------------------------------------------------------------.


def define_summary_log(list_logs):
    """Define station summary log file from list of file logs.

    It select only logged lines with root, WARNING and ERROR keywords.
    It write the summary log in the parent directory.
    """
    list_logs = sorted(list_logs)
    logs_dir = os.path.dirname(list_logs[0])
    station_name = logs_dir.split(os.path.sep)[-1]
    summary_logs_dir = os.path.dirname(logs_dir)
    ####-----------------------------------------------------------------------.
    #### Define summary and problem logs
    # Define summary logs file name
    summary_fpath = os.path.join(summary_logs_dir, f"logs_summary_{station_name}.log")
    # Define logs keywords to select lines to copy into the summary log file
    # -- > "has started" and "has ended" is used to copy the line with the filename being processsed
    list_keywords = ["has started", "has ended", "WARNING", "ERROR"]  # "DEBUG"
    re_keyword = re.compile("|".join(list_keywords))
    # Filter and concat all logs files
    with open(summary_fpath, "w") as output_file:
        for log_fpath in list_logs:
            with open(log_fpath) as input_file:
                for line in input_file:
                    if re_keyword.search(line):
                        # Write line to output file
                        output_file.write(line)
    ####-----------------------------------------------------------------------.
    #### Define problem logs
    # Define problem logs file name
    problem_fpath = os.path.join(summary_logs_dir, f"logs_problem_{station_name}.log")
    # - Copy the log of files with warnings and error
    list_keywords = ["ERROR"]  # "WARNING"
    re_keyword = re.compile("|".join(list_keywords))
    any_problem = False
    with open(problem_fpath, "w") as output_file:
        for log_fpath in list_logs:
            log_with_problem = False
            # Check if a warning or error is reported
            with open(log_fpath) as input_file:
                for line in input_file:
                    if re_keyword.search(line):
                        log_with_problem = True
                        any_problem = True
                        break
            # If it is reported, copy the log file in the logs_problem file
            if log_with_problem:
                with open(log_fpath) as input_file:
                    output_file.write(input_file.read())

    # If no problems occured, remove the logs_problem_<station_name>.log file
    if not any_problem:
        os.remove(problem_fpath)
    return None


####---------------------------------------------------------------------------.


def log_debug(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include debug entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.debug(msg)
    if verbose:
        print(" - " + msg)


def log_info(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include info entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.info(msg)
    if verbose:
        print(" - " + msg)


def log_warning(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include warning entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.warning(msg)
    if verbose:
        print(" - " + msg)


def log_error(logger: logger, msg: str, verbose: bool = False) -> None:
    """Include error entry into log.

    Parameters
    ----------
    logger : logger
        Log object.
    msg : str
        Message.
    verbose : bool, optional
        Whether to verbose the processing.
        The default is False.
    """
    logger.error(msg)
    if verbose:
        print(" - " + msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
from disdrodb.l0 import run_l0a
from disdrodb.l0.l0_reader import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    station_name,
    # Processing options
    force=False,
    verbose=False,
    parallel=False,
    debugging_mode=False,
):
    ##------------------------------------------------------------------------.
    #### - Define column names
    column_names = ["time", "unknown", "raw_drop_number"]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = "   "
    # - Skip first row as columns names
    reader_kwargs["header"] = None
    # - Skip file with encoding errors
    reader_kwargs["encoding_errors"] = "ignore"
    # - Need for zipped raw file (GPM files)
    reader_kwargs["zipped"] = True
    # - Searched file into tar files
    reader_kwargs["file_name_to_read_zipped"] = "spectrum.txt"
    # - Avoid first column to become df index
    reader_kwargs["index_col"] = False
    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"
    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"
    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"
    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error", "-.-"]

    ##------------------------------------------------------------------------.
    #### - Define dataframe sanitizer function for L0 processing
    def df_sanitizer_fun(df):
        # - Import pandas
        import pandas as pd

        # - Convert time column to datetime
        try:
            df["time"] = pd.to_datetime(
                df["time"], format="%Y %m %d %H %M %S", errors="coerce"
            )
        except ValueError:
            df["time"] = pd.to_datetime(
                df["time"], format="%Y-%m-%d %H:%M:%S", errors="coerce"
            )

        # - Drop columns not agreeing with DISDRODB L0 standards
        df = df.drop(columns=["unknown"])

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_name>
    glob_patterns = "*.tar"

    ####----------------------------------------------------------------------.
    #### - Create L0A products
    run_l0a(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        station_name=station_name,
        # Custom arguments of the reader for L0A processing
        glob_patterns=glob_patterns,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
        # Processing options
        force=force,
        verbose=verbose,
        parallel=parallel,
        debugging_mode=debugging_mode,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.


def _execute_cmd(cmd, raise_error=False):
    """Execute command in the terminal, streaming output in python console."""
    from subprocess import Popen, PIPE, CalledProcessError

    with Popen(cmd, shell=True, stdout=PIPE, bufsize=1, universal_newlines=True) as p:
        for line in p.stdout:
            print(line, end="")

    # Raise error if command didn't run successfully
    if p.returncode != 0 and raise_error:
        raise CalledProcessError(p.returncode, p.args)


def parse_arg_to_list(args):
    """Utility to pass list to command line scripts.

    If args = '' --> None
    If args = 'variable' --> [variable]
    If args = 'variable1 variable2' --> [variable1, variable2]
    """
    # If '', set to None
    args = None if args == "" else args
    # - If multiple arguments, split by space
    if isinstance(args, str):
        # - Split by space
        list_args = args.split(" ")
        # - Remove '' (deal with multi space)
        args = [args for args in list_args if len(args) > 0]
    return args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

import os
import logging

logger = logging.getLogger(__name__)


####--------------------------------------------------------------------------.


def _get_readers_directory() -> str:
    """Returns the path to the disdrodb.l0.readers directory within the disdrodb package."""
    # Current file path
    l0_folder_path = os.path.dirname(__file__)

    # Readers folder path
    reader_folder_path = os.path.join(l0_folder_path, "readers")
    return reader_folder_path


def _get_readers_data_sources() -> list:
    """Returns the readers data sources available at disdrodb.l0.readers"""
    # Readers folder path
    reader_folder_path = _get_readers_directory()

    # List of readers folder
    list_data_sources = [
        os.path.basename(f.path) for f in os.scandir(reader_folder_path) if f.is_dir()
    ]
    # Directory to remove
    bad_dirs = ["__pycache__", ".ipynb_checkpoints"]
    list_data_sources = [name for name in list_data_sources if name not in bad_dirs]

    return list_data_sources


def _get_readers_data_sources_path() -> list:
    """Returns the list of readers data sources directory paths within disdrodb.l0.readers"""
    # Readers folder path
    reader_folder_path = _get_readers_directory()

    # List of readers folder
    list_data_sources = [f.path for f in os.scandir(reader_folder_path) if f.is_dir()]
    return list_data_sources


def _get_readers_paths_by_data_source(data_source):
    """Return the filepath list of available readers for a specific data source.

    This function does not check the data_source validity.
    """
    # Retrieve reader data source directory
    reader_folder_path = _get_readers_directory()
    reader_data_source_path = os.path.join(reader_folder_path, data_source)
    if not os.path.isdir(reader_data_source_path):
        raise ValueError(f"No {data_source} directory in disdrodb.l0.readers")
    # Retrieve list of available readers paths
    list_readers_paths = [
        f.path
        for f in os.scandir(reader_data_source_path)
        if f.is_file() and f.path.endswith(".py")
    ]
    return list_readers_paths


def _get_readers_names_by_data_source(data_source):
    """Return the reader available for a given data_source.

    This function does not check the data_source validity.
    """
    # Retrieve reader data source directory
    list_readers_paths = _get_readers_paths_by_data_source(data_source)
    list_readers_names = [
        os.path.basename(path).replace(".py", "") for path in list_readers_paths
    ]
    return list_readers_names


####--------------------------------------------------------------------------.


def _check_reader_data_source(reader_data_source: str) -> str:
    """Check if the provided data source exists within the available readers.

    Please run get_available_readers_dict() to get the list of all available reader.

    Parameters
    ----------
    reader_data_source : str
        The directory within which the reader_name is located in the
        disdrodb.l0.readers directory.

    Returns
    -------
    str
        If data source is valid, return the correct data source name

    Raises
    ------
    ValueError
        Error if the data source name provided is not a directory within the disdrodb.l0.readers directory.
    """

    # List available readers data sources
    available_reader_data_sources = _get_readers_data_sources()
    # If not valid data_source, raise error
    if reader_data_source not in available_reader_data_sources:
        msg = f"Reader data source {reader_data_source} is not a directory inside the disdrodb.l0.readers directory."
        logger.error(msg)
        raise ValueError(msg)
    return reader_data_source


def check_reader_exists(reader_data_source: str, reader_name: str) -> str:
    """Check if the provided data source exists and reader names exists within the available readers.

    Please run get_available_readers_dict() to get the list of all available reader.

    Parameters
    ----------
    reader_data_source : str
        The directory within which the reader_name is located in the
        disdrodb.l0.readers directory.
    reader_name : str
        Campaign name

    Returns
    -------
    str
        If True : returns the reader name
        If False : Error - return None

    Raises
    ------
    ValueError
        Error if the reader name provided for the campaign has not been found.
    """
    # Check valid data_source
    reader_data_source = _check_reader_data_source(reader_data_source)
    # Get available reader names
    list_readers_names = _get_readers_names_by_data_source(reader_data_source)
    # If not valid reader_name, raise error
    if reader_name not in list_readers_names:
        msg = f"Reader {reader_name} is not valid. Valid readers {list_readers_names}."
        logger.exception(msg)
        raise ValueError(msg)
    return reader_name


def get_available_readers_dict() -> dict:
    """Returns the readers description included into the current release of DISDRODB.

    Returns
    -------
    dict
        The dictionary has the following schema {"data_source": {"reader_name": "reader_file_path"}}
    """
    # Format:
    # {data_source: {reader_name: reader_path,
    #                reader_name1: reader_path1}
    # }
    # Get list of reader data sources
    list_reader_data_sources = _get_readers_data_sources()

    # Build dictionary
    dict_reader = {}
    for data_source in list_reader_data_sources:
        # Retrieve the filepath of the available readers
        list_readers_paths = _get_readers_paths_by_data_source(data_source)
        # Initialize the data_source dictionary
        dict_reader[data_source] = {}
        for reader_path in list_readers_paths:
            reader_name = os.path.basename(reader_path).replace(".py", "")
            dict_reader[data_source][reader_name] = reader_path
    # Return available dictionary
    return dict_reader


def available_readers(data_sources=None, reader_path=False):
    """Retrieve available readers information."""
    # Get available readers dictionary
    dict_readers = get_available_readers_dict()
    # If data sources is not None, subset the dictionary
    if data_sources is not None:
        # Check valid data sources
        if isinstance(data_sources, str):
            data_sources = [data_sources]
        data_sources = [
            _check_reader_data_source(data_source) for data_source in data_sources
        ]
        # Create new dictionary
        dict_readers = {
            data_source: dict_readers[data_source] for data_source in data_sources
        }
    # If reader_path=False, provide {data_source: [list_reader_names]}
    if not reader_path:
        dict_readers = {
            data_source: list(dict_readers.keys())
            for data_source, dict_readers in dict_readers.items()
        }
    return dict_readers


####--------------------------------------------------------------------------.


def get_reader(reader_data_source: str, reader_name: str) -> object:
    """Returns the reader function based on input parameters.

    Parameters
    ----------
    reader_data_source : str
        The directory within which the reader_name is located in the
        disdrodb.l0.readers directory.
    reader_name : str
        The reader name.

    Returns
    -------
    object
        The reader() function

    """
    # Check data source and reader_name validity
    reader_data_source = _check_reader_data_source(reader_data_source)
    reader_name = check_reader_exists(
        reader_data_source=reader_data_source, reader_name=reader_name
    )
    # Retrive reader function
    if reader_name:
        full_name = f"disdrodb.l0.readers.{reader_data_source}.{reader_name}.reader"
        module_name, unit_name = full_name.rsplit(".", 1)
        my_reader = getattr(__import__(module_name, fromlist=[""]), unit_name)

    return my_reader


####--------------------------------------------------------------------------.
#### Checks for reader args


def _get_expected_reader_arguments():
    """Return a list with the expected reader arguments."""
    expected_arguments = [
        "raw_dir",
        "processed_dir",
        "station_name",
        "force",
        "verbose",
        "parallel",
        "debugging_mode",
    ]
    return expected_arguments


def check_reader_arguments(reader):
    """Check the reader have the expected input arguments."""
    import inspect

    signature = inspect.signature(reader)
    reader_arguments = sorted(list(signature.parameters.keys()))
    expected_arguments = sorted(_get_expected_reader_arguments())
    if reader_arguments != expected_arguments:
        raise ValueError(
            f"The reader must be defined with the following arguments: {expected_arguments}"
        )
    return None


####--------------------------------------------------------------------------.
#### Checks for metadata reader key


def _check_metadata_reader(metadata):
    """Check reader key is available and there is the associated reader."""
    # Check the reader is specified
    if "reader" not in metadata:
        raise ValueError("The reader is not specified in the metadata.")
    # If the reader name is specified, test it is valid.
    # - Convention: reader: "<data_source>/<reader_name>" in disdrodb.l0.readers
    reader_reference = metadata.get("reader")
    # - Check it contains /
    if "/" not in reader_reference:
        raise ValueError(
            f"The reader '{reader_reference}' reported in the metadata is not valid. Must have '<data_source>/<reader_name>' pattern."
        )
    # - Get the reader_reference component list
    reader_components = reader_reference.split("/")
    # - Check composed by two elements
    if len(reader_components) != 2:
        raise ValueError(
            "Expecting the reader reference to be composed of <data_source>/<reader_name>."
        )
    # - Retrieve reader data source and reader name
    reader_data_source = reader_components[0]
    reader_name = reader_components[1]

    # - Check the reader is available
    check_reader_exists(reader_data_source=reader_data_source, reader_name=reader_name)

    return None


def get_reader_from_metadata_reader_key(reader_data_source_name):
    """Retrieve the reader from the `reader` metadata value.

    The convention for metadata reader key: <data_source/reader_name> in disdrodb.l0.readers
    """
    reader_data_source = reader_data_source_name.split("/")[0]
    reader_name = reader_data_source_name.split("/")[1]
    reader = get_reader(reader_data_source=reader_data_source, reader_name=reader_name)
    return reader


def _get_reader_from_metadata(metadata):
    """Retrieve the reader from the metadata key `reader`

    The convention for metadata reader key: <data_source/reader_name> in disdrodb.l0.readers
    """
    reader_data_source_name = metadata.get("reader")
    return get_reader_from_metadata_reader_key(reader_data_source_name)


def get_station_reader(disdrodb_dir, data_source, campaign_name, station_name):
    """Retrieve reader form station metadata information."""
    from disdrodb.api.io import get_metadata_dict

    # Get metadata
    metadata = get_metadata_dict(
        disdrodb_dir=disdrodb_dir,
        product_level="RAW",
        data_source=data_source,
        campaign_name=campaign_name,
        station_name=station_name,
    )
    # ------------------------------------------------------------------------.
    # Check reader key is within the dictionary
    if "reader" not in metadata:
        raise ValueError(
            f"The `reader` key is not available in the metadata of the {data_source} {campaign_name} {station_name} station."
        )

    # ------------------------------------------------------------------------.
    # Check reader name validity
    _check_metadata_reader(metadata)

    # ------------------------------------------------------------------------.
    # Retrieve reader
    reader = _get_reader_from_metadata(metadata)

    # ------------------------------------------------------------------------.
    # Check reader argument
    check_reader_arguments(reader)

    return reader


####--------------------------------------------------------------------------.
#### Readers Docs


def is_documented_by(original):
    """Wrapper function to apply generic docstring to the decorated function.

    Parameters
    ----------
    original : function
        Function to take the docstring from.
    """

    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def reader_generic_docstring():
    """Script to convert the raw data to L0A format.

    Parameters
    ----------
    raw_dir : str
        The directory path where all the raw content of a specific campaign is stored.
        The path must have the following structure:
            <...>/DISDRODB/Raw/<data_source>/<campaign_name>'.
        Inside the raw_dir directory, it is required to adopt the following structure:
        - /data/<station_name>/<raw_files>
        - /metadata/<station_name>.yaml
        Important points:
        - For each <station_name> there must be a corresponding YAML file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files.
        - The campaign_name are expected to be UPPER CASE.
    processed_dir : str
        The desired directory path for the processed DISDRODB L0A and L0B products.
        The path should have the following structure:
            <...>/DISDRODB/Processed/<data_source>/<campaign_name>'
        For testing purpose, this function exceptionally accept also a directory path simply ending
        with <campaign_name> (i.e. /tmp/<campaign_name>).
    station_name : str
        Station name
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is True.
    parallel : bool
        If True, the files are processed simultanously in multiple processes.
        The number of simultaneous processes can be customized using the dask.distributed LocalCluster.
        If False, the files are processed sequentially in a single process.
        If False, multi-threading is automatically exploited to speed up I/0 tasks.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        It processes just the first 3 raw data files.
        The default is False.
    """


####--------------------------------------------------------------------------.
#### Check DISDRODB readers


def check_available_readers():
    """Check the readers arguments of all package."""
    dict_all_readers = available_readers(data_sources=None, reader_path=False)
    for reader_data_source, list_reader_name in dict_all_readers.items():
        for reader_name in list_reader_name:
            try:
                reader = get_reader(
                    reader_data_source=reader_data_source, reader_name=reader_name
                )
                check_reader_arguments(reader)
            except Exception as e:
                raise ValueError(
                    f"Unvalid reader for {reader_data_source}/{reader_name}.py. The error is {e}"
                )
    return None


####--------------------------------------------------------------------------.
